/*
 * Licensed to Crate under one or more contributor license agreements.
 * See the NOTICE file distributed with this work for additional
 * information regarding copyright ownership.  Crate licenses this file
 * to you under the Apache License, Version 2.0 (the "License"); you may
 * not use this file except in compliance with the License.  You may
 * obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
 * implied.  See the License for the specific language governing
 * permissions and limitations under the License.
 *
 * However, if you have executed another commercial license agreement
 * with Crate these terms will supersede the license and you may use the
 * software solely pursuant to the terms of the relevant commercial
 * agreement.
 */

package io.crate.planner.node.management;

import io.crate.analyze.AnalyzedPromoteReplica;
import io.crate.analyze.AnalyzedRerouteAllocateReplicaShard;
import io.crate.analyze.AnalyzedRerouteCancelShard;
import io.crate.analyze.AnalyzedRerouteMoveShard;
import io.crate.analyze.AnalyzedStatement;
import io.crate.analyze.AnalyzedStatementVisitor;
import io.crate.analyze.PartitionPropertiesAnalyzer;
import io.crate.analyze.SymbolEvaluator;
import io.crate.common.annotations.VisibleForTesting;
import io.crate.common.collections.Lists2;
import io.crate.data.Row;
import io.crate.data.Row1;
import io.crate.data.RowConsumer;
import io.crate.execution.support.OneRowActionListener;
import io.crate.expression.symbol.Symbol;
import io.crate.metadata.CoordinatorTxnCtx;
import io.crate.metadata.Functions;
import io.crate.metadata.PartitionName;
import io.crate.metadata.doc.DocTableInfo;
import io.crate.metadata.table.ShardedTable;
import io.crate.planner.DependencyCarrier;
import io.crate.planner.Plan;
import io.crate.planner.PlannerContext;
import io.crate.planner.operators.SubQueryResults;
import io.crate.sql.tree.Assignment;
import io.crate.sql.tree.GenericProperties;
import io.crate.types.DataTypes;
import org.elasticsearch.action.admin.cluster.reroute.ClusterRerouteRequest;
import org.elasticsearch.cluster.node.DiscoveryNodes;
import org.elasticsearch.cluster.routing.allocation.command.AllocateReplicaAllocationCommand;
import org.elasticsearch.cluster.routing.allocation.command.AllocateStalePrimaryAllocationCommand;
import org.elasticsearch.cluster.routing.allocation.command.AllocationCommand;
import org.elasticsearch.cluster.routing.allocation.command.CancelAllocationCommand;
import org.elasticsearch.cluster.routing.allocation.command.MoveAllocationCommand;

import java.util.List;
import java.util.Locale;
import java.util.function.Function;

import static io.crate.planner.NodeSelection.resolveNodeId;

public class AlterTableReroutePlan implements Plan {

    private static final InnerVisitor REROUTE_STATEMENTS_VISITOR = new InnerVisitor();
    private final AnalyzedStatement rerouteStatement;

    public AlterTableReroutePlan(AnalyzedStatement rerouteStatement) {
        this.rerouteStatement = rerouteStatement;
    }

    @Override
    public StatementType type() {
        return StatementType.MANAGEMENT;
    }

    @Override
    public void executeOrFail(DependencyCarrier dependencies,
                              PlannerContext plannerContext,
                              RowConsumer consumer,
                              Row params,
                              SubQueryResults subQueryResults) {
        var rerouteCommand = createRerouteCommand(
            rerouteStatement,
            plannerContext.transactionContext(),
            plannerContext.functions(),
            params,
            subQueryResults,
            dependencies.clusterService().state().nodes());

        dependencies
            .transportActionProvider()
            .transportClusterRerouteAction()
            .execute(
                new ClusterRerouteRequest().add(rerouteCommand),
                new OneRowActionListener<>(
                    consumer, r -> new Row1(r == null ? -1L : 1L)));
    }

    @VisibleForTesting
    public static AllocationCommand createRerouteCommand(AnalyzedStatement reroute,
                                                         CoordinatorTxnCtx txnCtx,
                                                         Functions functions,
                                                         Row parameters,
                                                         SubQueryResults subQueryResults,
                                                         DiscoveryNodes nodes) {
        Function<? super Symbol, Object> eval = x -> SymbolEvaluator.evaluate(
            txnCtx,
            functions,
            x,
            parameters,
            subQueryResults
        );

        return reroute.accept(
            REROUTE_STATEMENTS_VISITOR,
            new Context(nodes, eval));
    }

    private static class Context {

        private final DiscoveryNodes nodes;
        private final Function<? super Symbol, Object> eval;

        Context(DiscoveryNodes nodes,
                Function<? super Symbol, Object> eval) {
            this.nodes = nodes;
            this.eval = eval;
        }
    }

    private static class InnerVisitor extends AnalyzedStatementVisitor<Context, AllocationCommand> {

        @Override
        protected AllocationCommand visitAnalyzedStatement(AnalyzedStatement analyzedStatement, Context context) {
            throw new UnsupportedOperationException(
                String.format(Locale.ENGLISH, "Can't handle \"%s\"", analyzedStatement));
        }

        @Override
        public AllocationCommand visitReroutePromoteReplica(AnalyzedPromoteReplica statement,
                                                            Context context) {
            var boundedPromoteReplica = statement.promoteReplica().map(context.eval);

            String index = getRerouteIndex(
                statement.shardedTable(),
                Lists2.map(statement.partitionProperties(), x -> x.map(context.eval)));
            String toNodeId = resolveNodeId(
                context.nodes,
                DataTypes.STRING.value(boundedPromoteReplica.node()));

            return new AllocateStalePrimaryAllocationCommand(
                index,
                DataTypes.INTEGER.value(boundedPromoteReplica.shardId()),
                toNodeId,
                DataTypes.BOOLEAN.value(context.eval.apply(statement.acceptDataLoss()))
            );

        }

        @Override
        protected AllocationCommand visitRerouteMoveShard(AnalyzedRerouteMoveShard statement,
                                                          Context context) {
            var boundedMoveShard = statement.rerouteMoveShard().map(context.eval);

            String index = getRerouteIndex(
                statement.shardedTable(),
                Lists2.map(statement.partitionProperties(), x -> x.map(context.eval)));
            String toNodeId = resolveNodeId(
                context.nodes,
                DataTypes.STRING.value(boundedMoveShard.toNodeIdOrName()));

            return new MoveAllocationCommand(
                index,
                DataTypes.INTEGER.value(boundedMoveShard.shardId()),
                DataTypes.STRING.value(boundedMoveShard.fromNodeIdOrName()),
                toNodeId
            );
        }

        @Override
        protected AllocationCommand visitRerouteAllocateReplicaShard(AnalyzedRerouteAllocateReplicaShard statement,
                                                                     Context context) {
            var boundedRerouteAllocateReplicaShard = statement
                .rerouteAllocateReplicaShard()
                .map(context.eval);

            String index = getRerouteIndex(
                statement.shardedTable(),
                Lists2.map(statement.partitionProperties(), x -> x.map(context.eval)));
            String toNodeId = resolveNodeId(
                context.nodes,
                DataTypes.STRING.value(boundedRerouteAllocateReplicaShard.nodeIdOrName()));

            return new AllocateReplicaAllocationCommand(
                index,
                DataTypes.INTEGER.value(boundedRerouteAllocateReplicaShard.shardId()),
                toNodeId
            );
        }

        @Override
        protected AllocationCommand visitRerouteCancelShard(AnalyzedRerouteCancelShard statement,
                                                            Context context) {
            var boundedRerouteCancelShard = statement
                .rerouteCancelShard()
                .map(context.eval);

            boolean allowPrimary = validateCancelRerouteProperty(
                "allow_primary", boundedRerouteCancelShard.properties());

            String index = getRerouteIndex(
                statement.shardedTable(),
                Lists2.map(statement.partitionProperties(), x -> x.map(context.eval)));
            String nodeId = resolveNodeId(
                context.nodes,
                DataTypes.STRING.value(boundedRerouteCancelShard.nodeIdOrName()));

            return new CancelAllocationCommand(
                index,
                DataTypes.INTEGER.value(boundedRerouteCancelShard.shardId()),
                nodeId,
                allowPrimary
            );
        }

        private static String getRerouteIndex(ShardedTable shardedTable,
                                              List<Assignment<Object>> partitionsProperties) {
            if (shardedTable instanceof DocTableInfo) {
                DocTableInfo docTableInfo = (DocTableInfo) shardedTable;

                String indexName = docTableInfo.ident().indexNameOrAlias();
                PartitionName partitionName = PartitionPropertiesAnalyzer
                    .createPartitionName(partitionsProperties, docTableInfo);
                if (partitionName != null) {
                    indexName = partitionName.asIndexName();
                } else if (docTableInfo.isPartitioned()) {
                    throw new IllegalArgumentException(
                        "table is partitioned however no partition clause has been specified");
                }

                return indexName;
            }

            // Table is a blob table
            assert shardedTable.concreteIndices().length == 1 : "table has to contain only 1 index name";
            return shardedTable.concreteIndices()[0];
        }

        private static boolean validateCancelRerouteProperty(String propertyKey,
                                                             GenericProperties<Object> properties) {
            for (String key : properties.keys()) {
                if (propertyKey.equals(key)) {
                    return DataTypes.BOOLEAN.value(properties.get(propertyKey));
                } else {
                    throw new IllegalArgumentException(
                        String.format(Locale.ENGLISH, "\"%s\" is not a valid setting for CANCEL SHARD", key));
                }
            }
            return false;
        }
    }
}
